import os

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]
source_suffix = ".rst"
master_doc = "index"
project = "sonnetsuiteshelper"
year = "2024"
author = "Alan Manning"
copyright = f"{year}, {author}"
version = release = "0.4.6"

pygments_style = "trac"
templates_path = ["."]
extlinks = {
    "issue": ("https://github.com/Alan-Manning/python-sonnetsuiteshelper/issues/%s", "#"),
    "pr": ("https://github.com/Alan-Manning/python-sonnetsuiteshelper/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# if not on_rtd:  # only set the theme if we are building docs locally
html_theme = "sphinx_rtd_theme"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = f"{project}-{version}"

autodoc_default_options = {"members": None, "undoc-members": None, "private-members": None, "show-inheritance": None}

napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_ivar = True
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_preprocess_types = True
# napoleon_type_aliases = None
napoleon_custom_sections = [("KwArgs", "Keyword Arguments")]

# old
# napoleon_numpy_docstring = True
# # napoleon_include_init_with_doc = True
# napoleon_use_ivar = True
# napoleon_use_param = False
# napoleon_use_rtype = True
# # napoleon_preprocess_types = False
# # napoleon_type_aliases = True
# napoleon_custom_sections = [("KwArgs", "Keyword Arguments")]



import os
import subprocess
from os.path import isfile
from sys import stdout
from typing import Dict


def analyze_local(
    project_name: str,
    sonnet_install_loc: str,
    display_analysis_info_live: bool = False,
    lossles: bool = False,
    abs_cache_none: bool = False,
    abs_cache_stop_restart: bool = False,
    abs_cache_multi_sweep: bool = False,
    abs_no_discrete: bool = False,
    sub_freq_Hz: int | None = None,
    param_file: str = "",
):
    """Send a file to the local Sonnet Suites Solver.

    Parameters
    ----------
    project_name : str
        The name of the sonnet file to be analyzed. If this does not include
        the ".son" file extention then it will be added.

    sonnet_install_loc : str
        This the directory of the sonnet install. This is needed to know the
        location of the em executable to be able to run the analysis.
        This is usually for windows:
        >>> C:/Program Files/Sonnet Software/XX.XX
        where 'XX.XX' is the version number, e.g. ".../Sonnet Software/17.56".

    display_analysis_info_live: bool
        Whether to display live analysis info in the terminal.

    lossles : bool
        Default False

    abs_cache_none : bool
        Default False,

    abs_cache_stop_restart: bool = False,
        Default False

    abs_cache_multi_sweep: bool = False,
        Default False

    abs_no_discrete: bool = False,
        Default False

    sub_freq_Hz: int | None = None,
        Default None

    param_file: str = "",
        Default is blank str

    See Also
    --------
    analyze_remote : remote server analysis.


    Referencees
    -----------
    .. _Sonnet User's Guide:
        https://www.sonnetsoftware.com/support/help-18/users_guide/Sonnet%20User's%20Guide.html?EmCommandLineforBatch.html
    """

    em_exec = os.path.join(sonnet_install_loc, r"bin\em")

    # Check the em executable exists for the provided sonnet install loc.
    if not os.path.isfile(em_exec):
        raise FileNotFoundError("Could not find em executable in sonnet install path.")

    # Check if the project name ends .son if not add it.
    if project_name[:-4] != ".son":
        project_name += ".son"

    # Check the project file exists.
    if not os.path.isfile(project_name):
        raise (FileNotFoundError)

    run_cmd = f"{em_exec} {project_name}"

    if display_analysis_info_live:
        run_cmd += " -v"

    if lossles:
        run_cmd += " -Lossles"

    if abs_cache_none:
        run_cmd += " -AbsCacheNone"

    if abs_cache_stop_restart:
        run_cmd += " -AbsCacheStopRestart"

    if abs_cache_multi_sweep:
        run_cmd += " -AbsCacheMultiSweep"

    if abs_no_discrete:
        run_cmd += " -AbsNoDiscrete"

    if sub_freq_Hz:
        run_cmd += f" -SubFreqHz[{sub_freq_Hz}]"

    if param_file:
        run_cmd += f" -ParamFile {param_file}"

    cmd_output = subprocess.Popen(run_cmd, shell=True, stdout=subprocess.PIPE.stdout.read())

    return


def analyze_remote(project_name: str, remote_host: str, remote_port: str, param_file: str = ""):
    """Send a file to a remote emsolver server for analysis.

    Parameters
    ----------
    project_name : str
        This is the name of the sonnet file to be analyzed. If this does not
        include the .son file extention it will be added.

    remote_host : str
        This is the host name for the remote solver. e.g. "10.1.10.30"

    remote_port : str
        This is the port to be used to connect to the remote solver. e.g. "56150"

    KwArgs
    ------
    param_file: str
        This is a parameter file name for a project. This should include the
        path for the file and the file extention.
    """
    run_cmd = ""

    # get emclient_path in argument
    emclient_path = r'"C:\Program Files\Sonnet Software\17.56\bin\emclient"'

    shell = "TODO"

    if shell == "ps":
        run_cmd += "& "

    run_cmd = emclient_path

    # If this should be run remote try to add the remote to the run command
    # remote = True
    # if remote:
    #     keys_needed = ["host", "port"]
    #     if all(key in remote for key in keys_needed):
    #         run_cmd += f' -Server {remote["host"]}:{remote["port"]}'
    #     else:
    #         raise (
    #             KeyError(
    #                 f"remote parameter does not contain the keys needed.\nThe keys needed are {keys_needed}.\nCurrent keys are {list(remote.keys())}"
    #             )
    #         )

    run_cmd += f" -ProjectName {project_name}"

    # If this should be run with a parameter file try to add it to the run command
    if param_file:
        if os.path.isfile(param_file):
            run_cmd += f" -ParamFile {param_file}"
        else:
            raise FileNotFoundError("Could not find the parameter file in given directory")

    run_cmd += r" -Analyze"

    # run the command that has been built up and capture output
    cmd_output = subprocess.Popen(run_cmd, shell=True, stdout=subprocess.PIPE).stdout.read()

    return cmd_output



import os

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy.interpolate import UnivariateSpline
from scipy.optimize import curve_fit
from scipy.signal import find_peaks


def S21_mag_fit(freqs, f0, qr, qc):
    """Fit for the S21 mag against frequency."""
    xr = freqs / f0 - 1
    s21 = 1 - ((qr / qc) * (1 / (1 + 2j * qr * xr)))
    return np.abs(s21)


def volts_to_db(S21_volts):
    """Transforming power S21 volts**2 into decibells."""
    return 20 * np.log10(S21_volts)


class SonnetCSVOutputFile:
    """Csv output object.

    Loads in a csv output generated from a sonnet and provides functions to
    analyse that data.

    Parameters
    ----------
    file_name : str
        The name of the file.

    file_path : str
        The path of the file.

    Parameter : str
        The parameter type of the csv output file from sonnet. Default is "S-Param".

    complex : str
        the complex type of the csv output file from sonnet. Default is "Real-Imag".
    """

    def __init__(self, file_name: str, file_path: str = "", parameter: str = "S-Param", complex: str = "Real-Imag"):
        # if no .csv file extention it is added.
        if file_name[-4:] != ".csv":
            file_name = file_name + ".csv"

        self.file_name = file_name[:-4]

        live_file = os.path.join(file_path, file_name)

        file_exists = os.path.isfile(live_file)
        if not file_exists:
            raise (FileNotFoundError)

        # read csv

        csv_col_names = ["Frequency (GHz)", "RE[S11]", "IM[S11]", "RE[S12]", "IM[S12]", "RE[S21]", "IM[S21]", "RE[S22]", "IM[S22]"]

        # itterate to find start of the file's data
        skip_row = 2
        file_data = pd.read_csv(live_file, names=csv_col_names, skiprows=skip_row)

        while isinstance(file_data["Frequency (GHz)"][0], str):
            file_data = pd.read_csv(live_file, names=csv_col_names, skiprows=skip_row)
            skip_row += 1

        self.file_data = file_data
        self.freqs = np.array(self.file_data["Frequency (GHz)"] * 1e9)  # converty GHz to Hz
        self.S21_mag = np.array(np.abs(file_data["RE[S21]"] + 1j * file_data["IM[S21]"]))
        self.S21_mag_dB = volts_to_db(self.S21_mag)

    def __str__(self):
        return f"SonnetCSVOutputFile\n\tname: {self.file_name}\n\tParameter: {self.parameter}\n\tComplex: {self.complex}"

    def _get_indices_around_peak(self, y_data: list, no_points_around_peak: int = 200) -> list:
        """Get the indices around the peak in the data."""
        peaks_in_data = find_peaks(y_data, height=5, distance=100)
        if len(peaks_in_data[0]) == 0:
            raise (Exception("No Peaks found in data"))

        peak_index = peaks_in_data[0][0]

        # Get the range around the peak being sure to not overflow the original
        # array by indexing outside its length
        lower_range_index = max([peak_index - no_points_around_peak, 0])
        upper_range_index = min([peak_index + no_points_around_peak, len(y_data)])

        indices_around_peak = range(lower_range_index, upper_range_index)
        indices_around_peak = range(peak_index - no_points_around_peak, peak_index + no_points_around_peak)

        return indices_around_peak

    def plot_data(
        self, x_ax: str = "freq_MHz", y_ax: str = "S21_mag_dB", data_points_around_peak: int = 0, fig_ax: plt.Axes = None
    ) -> None:
        """Plots the data in the csv, default is plotting all data in the
        S21_mag_dB against freq. This can take differing x_ax and y_ax values
        to plot. Can also just plot region around the peak in the data by
        passing a data_points_around_peak value.

        KwArgs
        ----------
        x_ax : str
            The data to plot on the x axis. Default is the frequency in units
            of MHz. This parameter can take any of the values,
            ["freq_Hz", "freq_MHz", "freq_GHz"].

        y_ax : str
            The data to plot on the y axis. Default is the S21 magnitude in
            units of decibells. This parameter can take any of the values,
            ["S21_mag", "S21_mag_dB"].

        data_points_around_peak : int
            The number of data points either side of the peak in the data to
            plot. By default this value is 0 which will plot all the data.

        fig_ax : plt.Axes
            This is a matplotlib axes which, when defined, will be the axes the
            data is plotted to. This allows for customizing the look and adding
            extra data to the plot.
        """

        x_lookup = {
            "freq_Hz": {"data": self.freqs, "label": "freq", "units": "Hz"},
            "freq_MHz": {"data": self.freqs * 1e-6, "label": "freq", "units": "MHz"},
            "freq_GHz": {"data": self.freqs * 1e-9, "label": "freq", "units": "GHz"},
        }

        y_lookup = {
            "S21_mag": {"data": self.S21_mag, "label": "S21 mag", "units": "V**2"},
            "S21_mag_dB": {"data": volts_to_db(self.S21_mag), "label": "S21 mag", "units": "dB"},
        }

        if x_ax not in x_lookup.keys():
            raise KeyError("{x_ax} is an invalid value for x_ax. Valid values are {list(x_lookup.keys())}")

        if y_ax not in y_lookup.keys():
            raise KeyError("{y_ax} is an invalid value for y_ax. Valid values are {list(y_lookup.keys())}")

        x_data = x_lookup[x_ax]["data"]
        x_label = x_lookup[x_ax]["label"]
        x_units = x_lookup[x_ax]["units"]

        y_data = y_lookup[y_ax]["data"]
        y_label = y_lookup[y_ax]["label"]
        y_units = y_lookup[y_ax]["units"]

        title = f"File = {self.file_name}"
        col = "C0"

        if data_points_around_peak != 0:
            region = self._get_indices_around_peak(-self.S21_mag_dB, no_points_around_peak=data_points_around_peak)
            y_data = y_lookup[y_ax]["data"][region]
            x_data = x_lookup[x_ax]["data"][region]
            title += f"   ({data_points_around_peak}_points_around_peak)"
            col = "C1"

        # If a matplotlib figure axes is defined just plot the data to that.
        if fig_ax:
            fig_ax.scatter(x_data, y_data, s=0.5, color=col)
            fig_ax.plot(x_data, y_data, linewidth=0.2, alpha=0.3, color=col)
            fig_ax.set_xlabel(f"{x_label}     ({x_units})")
            fig_ax.set_ylabel(f"{y_label}     ({y_units})")
            return

        fig = plt.figure(title)
        rows = 1
        cols = 1
        grid = plt.GridSpec(rows, cols)  # , top=0.95, bottom=0.092, left=0.05, right=0.95, hspace=0.0, wspace=0.2)

        ax0 = plt.subplot(grid[0, 0])

        ax0.scatter(x_data, y_data, s=0.5, color=col)
        ax0.plot(x_data, y_data, linewidth=0.2, alpha=0.3, color=col)

        ax0.set_title(title, loc="left")
        ax0.set_xlabel(f"{x_label}     ({x_units})")
        ax0.set_ylabel(f"{y_label}     ({y_units})")
        ax0.grid(alpha=0.3)

        fig.show()

    def get_resonant_freq(self) -> float:
        """Get the resonant frequency (*in Hz*) from the data.

        Returns
        -------
        resonant_freq : float
            The resonant frequency of the peak in the data.
        """
        # find the peak in the data
        indices_around_peak = self._get_indices_around_peak(-self.S21_mag_dB)

        # Get the freqs and S21_mag around the peak
        freqs_around_peak = self.freqs[indices_around_peak]
        S21_mag_dB_around_peak = self.S21_mag_dB[indices_around_peak]

        # take the freq at the lowest point in the peak
        resonant_freq = freqs_around_peak[S21_mag_dB_around_peak.argmin()]

        return resonant_freq

    def get_Q_values(self) -> list:
        """Get the Q values from the data. Returns a list containing the QR, QC
        and QI values.

        Returns
        -------
        Q_Values : list
            list containing [QR, QC, QI].
        """

        # find the peak in the data
        indices_around_peak = self._get_indices_around_peak(-self.S21_mag_dB)

        # Get the freqs and S21_mag around the peak
        freqs_around_peak = self.freqs[indices_around_peak]
        S21_mag_around_peak = self.S21_mag[indices_around_peak]

        init_QR_guess = 10e3
        init_QC_guess = 2 * init_QR_guess
        init_guesses = np.array([self.get_resonant_freq(), init_QR_guess, init_QC_guess])
        popt, pcov = curve_fit(S21_mag_fit, freqs_around_peak, S21_mag_around_peak, p0=init_guesses)
        QR = popt[1]
        QC = popt[2]
        QI = 1 / ((1 / QR) - (1 / QC))

        Q_Values = [QR, QC, QI]

        return Q_Values

    def get_three_dB_BW(self) -> float:
        """Get the 3dB BW from the peak in the data."""
        spline = UnivariateSpline(self.freqs, volts_to_db(self.S21_mag) + 3.0, s=0)
        return abs(spline.roots()[1] - spline.roots()[0])



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -msonnetsuiteshelper` python will execute
    ``__main__.py`` as a script. That means there will not be any
    ``sonnetsuiteshelper.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there"s no ``sonnetsuiteshelper.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import sys


def main(argv=sys.argv):
    """
    Args:
        argv (list): List of arguments

    Returns:
        int: A return code

    Does stuff.
    """
    print(argv)
    return 0



import os

import numpy as np
import yaml
from matplotlib import pyplot as plt

from . import analysis_tools, file_generation


class SimpleSingleParamOptimiser:
    """Optimiser for Sonnet files.

    This takes in an initial .son file and .csv output from running
    that sonnet file and will try to optimise that file for a specific value
    within some defined tolerance.

    Attributes
    ----------
    name : str
        The unique name of this optimisation used for file generation. If
        this conflicts with another optimisers name then there will be
        files overwriting eachother.

    batch_1_son_filename : str
        The name of the first sonnet file which is the starting point for
        the optimiser. e.g. "son_sim_V1.son"

    batch_1_son_file_path : str
        The file path of the first sonnet file which is the starting point
        for the optimiser. e.g. if the folder structure looks like this:
        "batch_1_son_files/son_sim_V1.son", then the path is "batch_1_son_files".

    batch_1_output_filename : str
        The name of the first output file from a simulation which is the
        starting point for the optimiser. e.g. "son_sim_V1.csv". Often this is
        the same name as the initial son file.

    batch_1_output_file_path : str
        The file path of the first output file which is the starting point for
        the optimiser. e.g. if the folder structure looks like this:
        "batch_1_out/son_sim_V1.csv", then the path is "batch_1_out".
        Often this is the same path as the son file path.

    correlation : int
        The correlation between the variable parameter and the desired
        output param. This accepts str values "+" or "-".
        If an increase in the variable_param_value results in an increase in the
        desired_output_param, then this should be "+".
        If an increase in the variable_param_value results in a decrease in the
        desired_output_param, Then this should be "-".

    desired_output_param : str
        The variable output the optimiser should be optimising for. e.g.
        if trying to get a resonant frequency of 2GHz, then this argument
        will be ' desired_output_param="f0" '.
        This will only take values of:
        {"QR", "QC", "QI", "f0", "three_dB_BW"}.

    desired_output_param_value : str
        The value of the desired_output_param. e.g. if trying to
        resonant frequency of 2GHz, then this argument will be
        'desired_output_param_value=2.0e9'.

    desired_output_param_value_tolerence_percent : float
        This is the percentage tolerance around the desired_output_param_value.
        eg. if trying to get resonant frequency of 2GHz +- 1%, then this
        argument will be 'desired_output_param_value_tolerence_percent = 0.01'.

    desired_output_param_values : list
        This is a list of the desired ouptut parameter values from the optimser.

    loaded_cache : bool
        This is a bool that shows whether a cache file has been used to laod
        values.

    sonnet_mesh_size : float
        Default=1.0. The mesh size in sonnet. This is the smallest change
        possible in the varaible parameter that will result in a different
        file being produced.

    next_variable_param_value : float
        This is a the next variable parameter value to be used in the optimser.

    variable_param_name : str
        The name of the parameter that should be varied by the optimiser
        to achieve the desired result.

    variable_param_values : list
        This is a list of the variable parameter values used in the optimser.
    """

    def __init__(
        self,
        unique_name: str,
        varaible_param_name: str,
        batch_1_son_filename: str,
        batch_1_son_file_path: str,
        batch_1_output_filename: str,
        batch_1_output_file_path: str,
        init_variable_param_value: float,
        desired_output_param: str,
        desired_output_param_value: float,
        desired_output_param_value_tolerence_percent: float,
        correlation: str,
        sonnet_mesh_size: float = 1.0,
        ignore_loading_cache: bool = False,
    ):
        """
        Parameters
        ----------
        unique_name : str
            The unique name of this optimisation used for file generation. If
            this conflicts with another optimisers name then there will be
            files overwriting eachother.

        varaible_param_name: str
            The name of the parameter that should be varied by the optimiser
            to achieve the desired result.

        batch_1_son_filename: str
            The name of the first sonnet file which is the starting point for
            the optimiser. e.g. "son_sim_V1.son"

        batch_1_son_file_path: str
            The file path of the first sonnet file which is the starting point
            for the optimiser. e.g. if the folder structure looks like this:
            "batch_1_son_files/son_sim_V1.son", then the path is "batch_1_son_files".

        batch_1_output_filename: str
            The name of the first output file from a simulation which is the
            starting point for the optimiser. e.g. "son_sim_V1.csv".
            Often this is the same name as the initial son file.

        batch_1_output_file_path: str
            The file path of the first output file which is the starting point for
            the optimiser. e.g. if the folder structure looks like this:
            "batch_1_out/son_sim_V1.csv", then the path is "batch_1_out".
            Often this is the same path as the son file path.

        desired_output_param : str
            The variable output the optimiser should be optimising for. e.g.
            if trying to get a resonant frequency of 2GHz, then this argument
            will be ' desired_output_param="f0" '.
            This will only take values of:
            {"QR", "QC", "QI", "f0", "three_dB_BW"}.

        desired_output_param_value: float
            The value of the desired_output_param. e.g. if trying to
            resonant frequency of 2GHz, then this argument will be
            'desired_output_param_value=2.0e9'.

        desired_output_param_value_tolerence_percent: float
            This is the percentage tolerance around the desired_output_param_value.
            eg. if trying to get resonant frequency of 2GHz +- 1%, then this
            argument will be 'desired_output_param_value_tolerence_percent = 0.01'.

        correlation: str
            The correlation between the variable parameter and the desired
            output param. This accepts str values "+" or "-".
            If an increase in the variable_param_value results in an increase
            in the desired_output_param, then this should be "+".
            If an increase in the variable_param_value results in a decrease
            in the desired_output_param, Then this should be "-".

        sonnet_mesh_size : float or int
            Default=1.0. The mesh size in sonnet. This is the smallest change
            possible in the varaible parameter that will result in a different
            file being produced.

        ignore_loading_cache : bool
            Default=False. When True the optimser will not try to load cached
            results into memory. Otherwise by default the optimser will try to
            load values from a cache file in the OptCache path.

        """

        # Check corrent format of arguments
        acceptable_output_param_strings = ["QR", "QC", "QI", "f0", "three_dB_BW"]
        if desired_output_param not in acceptable_output_param_strings:
            raise ValueError(f"Cannot optimise for {desired_output_param}. Can only optimise for {acceptable_output_param_strings}")

        acceptable_correlation_strings = ["+", "-"]
        if correlation not in acceptable_correlation_strings:
            raise ValueError(f"Cannot correlate {correlation}. Can only use {acceptable_correlation_strings}")

        if batch_1_son_filename[-4:] != ".son":
            raise ValueError(f"batch_1_son_filename should be a sonnet file. This means it should contain a '.son' file extention.")

        if batch_1_output_filename[-4:] != ".csv":
            raise ValueError(
                f"batch_1_output_filename should be a csv output file from a simulation. This means it should contain a '.csv' file extention."
            )
        # General setup
        self.name = unique_name

        # Check for an existing cache file if ignore_loading_cache is False
        if not ignore_loading_cache:
            self.load_cached_results()
        else:
            self.loaded_cache = False

        if self.loaded_cache:
            return

        self.correlation = +1 if correlation == "+" else -1

        # Variable param setup
        self.variable_param_name = varaible_param_name
        self.variable_param_values: list[float] = []

        # Desired param setup
        self.desired_output_param = desired_output_param
        self.desired_output_param_value = desired_output_param_value
        self.desired_output_param_value_tolerence_percent = desired_output_param_value_tolerence_percent
        self.desired_output_param_values: list[float] = []

        # File settings
        self.sonnet_mesh_size = sonnet_mesh_size

        # Analyse first batch initial simulations
        self.batch_1_son_filename = batch_1_son_filename
        self.batch_1_son_file_path = batch_1_son_file_path

        self.batch_1_output_filename = batch_1_output_filename
        self.batch_1_output_file_path = batch_1_output_file_path

        self.next_variable_param_value = init_variable_param_value
        self.analyze_batch()
        self.generate_next_batch()

    def __str__(self) -> str:
        string = "Simple_single_param_optimiser"
        string += f"\n\t   current_batch_no: {self.get_current_batch_no()}"
        string += f"\n\tvariable_param_name: {self.variable_param_name}"
        string += f"\n\tdesired_output_param: {self.desired_output_param}"
        string += f"\n\tdesired_output_param_value: {self.desired_output_param_value}"
        return string

    def _create_yaml_str(self) -> str:
        """Create a yaml formated string containing all the relevant instance
        variables."""
        state = dict(
            name=self.name,
            correlation=self.correlation,
            variable_param_name=self.variable_param_name,
            variable_param_values=self.variable_param_values,
            desired_output_param=self.desired_output_param,
            desired_output_param_value=self.desired_output_param_value,
            desired_output_param_value_tolerence_percent=self.desired_output_param_value_tolerence_percent,
            desired_output_param_values=self.desired_output_param_values,
            next_variable_param_value=self.next_variable_param_value,
            sonnet_mesh_size=self.sonnet_mesh_size,
            batch_1_son_filename=self.batch_1_son_filename,
            batch_1_son_file_path=self.batch_1_son_file_path,
            batch_1_output_filename=self.batch_1_output_filename,
            batch_1_output_file_path=self.batch_1_output_file_path,
        )

        state_str = "---\n"
        for key, val in state.items():
            if isinstance(val, str):
                state_str = state_str + f"{key}: '{val}'\n"
            else:
                state_str = state_str + f"{key}: {val}\n"

        return state_str

    def get_cache_filename_and_path(self) -> str:
        """Get the filename and file path for the optimiser cache file."""
        return os.path.join(self.get_cache_file_path(), self.get_cache_filename())

    def get_cache_file_path(self) -> str:
        """Get the file path for the optimiser cache file."""
        file_path = "OptCache"
        return file_path

    def get_cache_filename(self) -> str:
        """Get the filename for the optimiser cache file."""
        filename = f"SSPOC_{self.name}.yml"
        return filename

    def cache_results(self) -> None:
        """Cache the results of the optimiser so far into a yaml file.

        This results in the optimser not regernerating and reanalysing
        files.
        """
        # check output_directory exists and if not try to create it.
        if not os.path.isdir(self.get_cache_file_path()):
            try:
                os.mkdir(self.get_cache_file_path())
            except Exception as err:
                raise err

        yaml_str = self._create_yaml_str()
        with open(self.get_cache_filename_and_path(), "w+") as yaml_file:
            yaml_file.write(yaml_str)

        return

    def load_cached_results(self) -> None:
        """Load cached results of the optimiser so far if a cache file
        exists."""

        cache_file = self.get_cache_filename_and_path()

        if not os.path.isfile(cache_file):
            self.loaded_cache = False
            return

        try:
            with open(cache_file, "r") as stream:
                cached_data = yaml.safe_load(stream)

            self.name = cached_data.get("name")
            self.correlation = cached_data.get("correlation")
            self.variable_param_name = cached_data.get("variable_param_name")
            self.variable_param_values = cached_data.get("variable_param_values")
            self.desired_output_param = cached_data.get("desired_output_param")
            self.desired_output_param_value = cached_data.get("desired_output_param_value")
            self.desired_output_param_value_tolerence_percent = cached_data.get("desired_output_param_value_tolerence_percent")
            self.desired_output_param_values = cached_data.get("desired_output_param_values")
            self.next_variable_param_value = cached_data.get("next_variable_param_value")
            self.sonnet_mesh_size = cached_data.get("sonnet_mesh_size")
            self.batch_1_son_filename = cached_data.get("batch_1_son_filename")
            self.batch_1_son_file_path = cached_data.get("batch_1_son_file_path")
            self.batch_1_output_filename = cached_data.get("batch_1_output_filename")
            self.batch_1_output_file_path = cached_data.get("batch_1_output_file_path")

            self.loaded_cache = True
        except Exception as e:
            print("Error loading cache")
            raise e

        return

    def get_current_batch_no(self) -> int:
        """Get the current batch number."""
        current_batch_no = len(self.variable_param_values)
        return current_batch_no

    def get_next_batch_no(self) -> int:
        """Get the current batch number."""
        next_batch_number = self.get_current_batch_no() + 1
        return next_batch_number

    def get_next_output_filename(self) -> str:
        """Get the output filename for the next batch's generated file."""
        output_filename = f"batch_{self.get_next_batch_no()}__{self.name}_{self.variable_param_name}_{self.next_variable_param_value}.son"
        return output_filename

    def get_next_output_file_path(self) -> str:
        """Get the output filepath for the next batch's generated file."""
        output_filepath = f"batch_{self.get_next_batch_no()}_generated_files"
        return output_filepath

    def get_last_analysis_filename(self) -> str:
        """Get the filename of the output file from the last batch that has
        been simulated in sonnet."""
        # Strip the .son file extention and add csv extention
        analysed_filename = self.get_next_output_filename()[:-4] + ".csv"
        return analysed_filename

    def get_last_analysis_file_path(self) -> str:
        """Get the file path of the output file from the last batch that has
        been simulated in sonnet."""
        analysed_file_path = f"batch_{self.get_next_batch_no()}_outputs"
        return analysed_file_path

    def get_optimised_filename(self) -> str:
        """Get the filename of the optimised file."""
        # Strip the .son file extention and add csv extention
        if self.last_result_reached_optimisation():
            optimised_filename = self.get_last_output_filename()[:-4] + ".csv"
            return optimised_filename
        else:
            raise (LookupError("Could not get optimised filename. Optimiser has not yet found optimised values."))

    def get_optimised_file_path(self) -> str:
        """Get the file path of the optimised file."""
        if self.last_result_reached_optimisation():
            optimised_file_path = f"batch_{self.get_current_batch_no()}_outputs"
            return optimised_file_path
        else:
            raise (LookupError("Could not get optimised file path. Optimiser has not yet found optimised values."))

    def get_last_output_filename(self) -> str:
        """Get the output filename for the last batch's generated file."""
        if self.get_current_batch_no() == 1:
            output_filename = self.batch_1_son_filename
        else:
            output_filename = (
                f"batch_{self.get_current_batch_no()}__{self.name}_{self.variable_param_name}_{self.variable_param_values[-1]}.son"
            )
        return output_filename

    def get_last_output_file_path(self) -> str:
        """Get the output filepath for the next batch's generated file."""
        if self.get_current_batch_no() == 1:
            output_filepath = self.batch_1_son_file_path
        else:
            output_filepath = f"batch_{self.get_current_batch_no()}_generated_files"
        return output_filepath

    def get_last_desired_output_param_value(self) -> float:
        """Get the last desired output param value from the last analysis."""
        return self.desired_output_param_values[-1]

    def get_last_variable_param_value(self) -> float:
        """Get the last variable param value from the last analysis."""
        return self.variable_param_values[-1]

    def append_new_results_to_self(self, output_value: float) -> None:
        """Append the analysed results to self."""
        self.desired_output_param_values.append(output_value)
        self.variable_param_values.append(self.next_variable_param_value)
        return

    def last_result_reached_optimisation(self) -> bool:
        """Check of the last desired_output_param_value is within the tolerance
        of the desired_output_param_value.

        If des*(1+tol) <= (last output vlaue) <= des*(1+tol) -> True
        else False.
        """
        last = self.get_last_desired_output_param_value()
        des = self.desired_output_param_value
        tol = self.desired_output_param_value_tolerence_percent
        outside = False
        if last <= des * (1 - tol):
            outside = True

        if last >= des * (1 + tol):
            outside = True

        if outside:
            return False
        else:
            return True

    def round_to_sonnet_mesh_size(self, value) -> float:
        """Round the input value to the nearest sonnet_mesh_size step."""
        rounded_result = np.round(float(value) / self.sonnet_mesh_size) * self.sonnet_mesh_size
        return rounded_result

    def get_next_variable_param_value(self) -> float:
        """Generate the next varialbe param value for the next batch output
        file."""

        # If less than or 3 batches run then use a simple scale to generate next
        if self.get_current_batch_no() <= 3:
            next_variable_param_value = self.get_next_variable_param_value_from_percent_scale()
            return next_variable_param_value

        next_variable_param_value = self.get_next_variable_param_value_from_lin_fit()

        # If the next is the same as the last then try to percentage scale rather than lin fit.
        if next_variable_param_value == self.get_last_variable_param_value():
            next_variable_param_value = self.get_next_variable_param_value_from_percent_scale()
            # If the next value is still the same try to change the value by the sonnet_mesh_size.
            if next_variable_param_value == self.get_last_variable_param_value():
                next_variable_param_value = self.change_variable_param_value_by_mesh_step()

        return next_variable_param_value

    def change_variable_param_value_by_mesh_step(self) -> float:
        """Changes the curent variable_param_value by the sonnet file mesh
        size.

        This follows similar logic to the percent scale but only
        increases or decreases the value by the sonnet_mesh_size.
        """
        if self.get_last_desired_output_param_value() > self.desired_output_param_value:
            new_value = self.get_last_variable_param_value() + (self.correlation * self.sonnet_mesh_size)
        else:
            new_value = self.get_last_variable_param_value() - (self.correlation * self.sonnet_mesh_size)

        return new_value

    def get_next_variable_param_value_from_percent_scale(self) -> float:
        """Get the next variable value from simply adding/subtracting to the
        previous variable value by some adjust adjust strength multiplied by
        the delta between the last output value and the fnal desired output
        value. i.e. next = last +- (delta_from_desired*adjust_strength).

        This result is then rounded to the nearest sonnet mesh grid
        point.
        """
        delta_from_desired = abs(self.get_last_desired_output_param_value() - self.desired_output_param_value)
        adjust_strength = 0.002
        delta_in_variable_param = adjust_strength * delta_from_desired

        if self.get_last_desired_output_param_value() > self.desired_output_param_value:
            new_value = self.get_last_variable_param_value() - (self.correlation * delta_in_variable_param)
        else:
            new_value = self.get_last_variable_param_value() + (self.correlation * delta_in_variable_param)

        return self.round_to_sonnet_mesh_size(new_value)

    def get_next_variable_param_value_from_lin_fit(self) -> float:
        """Get the next variable value by lin fitting the data. This makes a
        linear fit and then gets the variable value that results in an
        intersection with the desired_output_param_value.

        This result is then rounded to the nearest sonnet mesh grid
        point.
        """
        x_data = np.array(self.variable_param_values)
        y_data = np.array(self.desired_output_param_values)
        poly_fit = np.polyfit(y_data, x_data, deg=1)
        poly_func = np.poly1d(poly_fit)
        new_value = poly_func(self.desired_output_param_value)

        return self.round_to_sonnet_mesh_size(new_value)

    def generate_next_batch(self, override_variable_param_value: float | None = None, ignore_automatic_stop: bool = False) -> None:
        """Generate the next batch of simulation files.

        Parameters
        ----------
        override_variable_param_value: float
            This is the value that will be used to generate the next batch
            file.

        ignore_automatic_stop: bool
            Default False. If True this will continue to generate the next
            batch even though it has reached the desired value.

        Outputs
        -------
        sonnet file: .son file
            This generates a .son file with the varable param changed.
            The file name for this file is :
            "{batch_no}_{name}__{varaible_param_name}_{variable_param_value}.son"
        """
        if self.last_result_reached_optimisation():
            print(f"Optimiser {self.name} Reached desired QR")
            print(f"\tBatch_number = {self.get_current_batch_no()}")
            print(f"\t{self.desired_output_param} = {self.get_last_desired_output_param_value()}")
            print(f"\t{self.variable_param_name} = {self.get_last_variable_param_value()}")

        if (not ignore_automatic_stop) and self.last_result_reached_optimisation():
            return

        if override_variable_param_value is None:
            variable_param_value = self.get_next_variable_param_value()
        else:
            variable_param_value = override_variable_param_value

        self.next_variable_param_value = variable_param_value

        name_for_output_file = self.get_next_output_filename()
        output_file_path = self.get_next_output_file_path()

        params_to_edit_dict = {self.variable_param_name: variable_param_value}

        last_filename = self.get_last_output_filename()
        last_file_path = self.get_last_output_file_path()

        file_generation.generate_file_like(
            base_filename=last_filename,
            base_file_path=last_file_path,
            output_filename=name_for_output_file,
            output_file_path=output_file_path,
            params_to_edit=params_to_edit_dict,
        )

        self.cache_results()
        return

    def analyze_batch(self) -> None:
        """Analyze the current batch of simulations that have been run."""
        if self.get_current_batch_no() == 0:
            # special case for first batch
            son_csv = analysis_tools.SonnetCSVOutputFile(self.batch_1_output_filename, file_path=self.batch_1_output_file_path)
        else:
            filename = self.get_last_analysis_filename()
            file_path = self.get_last_analysis_file_path()

            son_csv = analysis_tools.SonnetCSVOutputFile(filename, file_path=file_path)

        f0 = son_csv.get_resonant_freq()
        QR, QC, QI = son_csv.get_Q_values()
        three_dB_BW = son_csv.get_three_dB_BW()

        match self.desired_output_param:
            case "QR":
                self.append_new_results_to_self(QR)
            case "QC":
                self.append_new_results_to_self(QC)
            case "QI":
                self.append_new_results_to_self(QI)
            case "f0":
                self.append_new_results_to_self(f0)
            case "three_dB_BW":
                self.append_new_results_to_self(three_dB_BW)
            case _:
                print("ERROR")
                raise (ValueError(f"Error, cannot optimise for {self.desired_output_param}"))

        return

    def plot_optimisation(
        self,
        fig_ax: plt.Axes | None = None,
        plot_fit_function: bool = True,
        plot_next_batch_variable_value: bool = True,
        set_axis_labels: bool = True,
    ) -> None:
        """Plot the current results of the optimiser."""

        x_data = np.array(self.variable_param_values)
        x_label = self.variable_param_name

        y_data = np.array(self.desired_output_param_values)
        y_label = self.desired_output_param

        face_color_for_optimised = "#b4f7ab"
        face_color_for_not_optimised = "#fcc7bd"

        if fig_ax is None:
            title = f"{self.name} - {self.desired_output_param} vs {self.variable_param_name}"
            fig = plt.figure(title)
            rows = 1
            cols = 1
            grid = plt.GridSpec(rows, cols)
            ax = plt.subplot(grid[0, 0])
        else:
            ax = fig_ax

        phi = np.linspace(0, np.pi, len(x_data))
        rgb_cycle = (
            np.stack(
                (
                    np.cos(phi),
                    np.cos(phi + 2 * np.pi / 3),
                    np.cos(phi - 2 * np.pi / 3),
                )
            ).T
            + 1
        ) * 0.5

        for i, (x, y) in enumerate(zip(x_data, y_data)):
            ax.scatter(x, y, s=5, color=rgb_cycle[i])
            ax.annotate(f"{i+1}", (x, y))

        ax.axhline(y=self.desired_output_param_value)

        if plot_fit_function and self.get_current_batch_no() > 3:
            poly_fit = np.polyfit(x_data, y_data, deg=1)
            poly_func = np.poly1d(poly_fit)
            x_ax = np.linspace(min(x_data) * 0.999, max(x_data) * 1.001, 100)

            ax.plot(x_ax, poly_func(x_ax), label="fit")

        if plot_next_batch_variable_value and (not self.last_result_reached_optimisation()):
            next_variable_param_value = self.get_next_variable_param_value()
            ax.scatter(
                next_variable_param_value,
                self.desired_output_param_value,
                s=50,
                color="red",
                marker="x",
            )
            ax.annotate("N", (next_variable_param_value, self.desired_output_param_value))

        if self.last_result_reached_optimisation():
            ax.set_facecolor(face_color_for_optimised)
            ax.text(
                0.5,
                0.5,
                f"Finished at BATCH {self.get_current_batch_no()}",
                transform=ax.transAxes,
                color="gray",
                alpha=0.6,
                ha="center",
                va="center",
            )
        else:
            ax.set_facecolor(face_color_for_not_optimised)

        if set_axis_labels:
            ax.set_xlabel(f"{x_label}")
            ax.set_ylabel(f"{y_label}")

        if fig_ax is None:
            fig.show()

        return



import os
import re
from typing import Dict


def generate_file_like(
    base_filename: str,
    output_filename: str,
    base_file_path: str = "",
    output_file_path: str = "",
    output_filename_prefix: str = "",
    output_filename_suffix: str = "",
    params_to_edit: Dict = {},
    general_metals_to_edit: Dict = {},
    adaptive_sweeps_to_edit: Dict = {},
    linear_sweeps_to_edit: Dict = {},
) -> None:
    """This generates a Sonnet file like the base file specified. This will
    take in a series of values or other elements to further modify the Sonnet
    file to be produced.

    Parameters
    ----------
    base_filename : str
        This is the name of the base file, this will assume a ".son" file
        extention if this is not included in the base file name. This will be
        used to create the new Sonnet file. This can include the path of the
        file if it does not exist in the same directory as the python script.

    output_filename : str
        This is the name for the output file to be generated. If this does not
        have a ".son" file extention already, one will be added. If a file with
        this name already exists then its contents will be overwritten.

    KwArgs
    ------
    output_file_path : str
        This is the directory that the output file should be saved to.
        By default this is a blank string which will save the file in the same
        directory as the script. When specified the output file will be saved
        to this directory. If this directory does not already exist it will be
        created.

    output_filename_prefix : str
        This is a string to add to the beggining of the filename. Default is a
        blank string.

    output_filename_suffix : str
        This is a string to add to the end of the filename. Default is a
        blank string.

    params_to_edit : dict
        This is a dictionary that has keys of parameter names and values of the
        values those parameters should take. Note the values must be of type
        float or int. This can take any number of key and value pairs. If there
        is any parameter that isnt in the base file that exists in this dict
        then this will print an error and continue with the other parameter
        values. An example dict to be passed would take the form:
        >>> Params_to_edit = {
        ...     "Length_var_1" : 400,
        ...     "Length_var_2" : 250,
        ...     "Length_var_3" : 1975,
        ...     ...
        ... }

    general_metals_to_edit : dict
        This is a dictionary that has keys of general metal names and values of
        dictionarys with "Rdc", "Rrf", "Xdc", "Ls" key names and values with
        values for those keys. Note those keys must exist in the dict for each
        general metal to edit. Any number of general metals can be passed. If
        there is any metal in this dictionary that doesn't exist in the base
        file, or if that metal is there but has the wrong dict keys, this will
        print an error and continue with the other metals if there exist any.
        An example dict to be passed would take the form:
        >>> gen_mets_edits = {
        ...     "gen_met_1": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        ...     "gen_met_2": {"Rdc": 1e-08, "Rrf": 14e-8, "Xdc": 0, "Ls": 0.003},
        ...     ...
        ... }

    adaptive_sweeps_to_edit : dict
        This is a dictionary that has keys of "sweep_min", "sweep_max",
        "target_freqs" and values of the values for those keys.
        An example dict to be passed would take the form:
        >>> adaptive_sweep = {
        ...     "sweep_min" : 1.0,
        ...     "sweep_max" : 5.0,
        ...     "target_freqs" : 500,
        ... }

    linear_sweeps_to_edit : dict
        This is a dictionary that has keys of "sweep_min", "sweep_max",
        "step_size" and values of the values for those keys.
        An example dict to be passed would take the form:
        >>> linear_sweep = {
        ...     "sweep_min" : 1.0,
        ...     "sweep_max" : 5.0,
        ...     "step_size" : 0.1,
        ... }

    Warnings
    --------
    If a file with this name already exists then its contents will be
    overwritten!
    """

    # if the basefile has no .son file extention add it.
    if base_filename[-4:] != ".son":
        base_filename = f"{base_filename}.son"

    # add the path for the basefile
    base_file = os.path.join(base_file_path, base_filename)

    # check base_file exists
    if not os.path.isfile(base_file):
        raise FileNotFoundError(f"Unable to find file: {base_file}")

    # check output_directory exists and if not try to create it.
    if not os.path.isdir(output_file_path):
        try:
            os.mkdir(output_file_path)
        except Exception as err:
            raise err

    # if .son file extention exists remove it. It is added next.
    if output_filename[-4:] == ".son":
        output_filename = output_filename[:-4]

    output_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"

    final_file = os.path.join(output_file_path, output_filename)

    with open(base_file) as f:
        contents = f.read()

    # If params dict not empty then change vals
    if params_to_edit:
        for key, val in params_to_edit.items():
            pattern = re.compile(rf'VALVAR {key} LNG (\w+|"\w+"|\d+\.\d+) "Dim\. Param\."')

            if pattern.search(contents):
                replacement = f'VALVAR {key} LNG {val} "Dim. Param."'
                contents = pattern.sub(replacement, contents)
            else:
                print(f"WARNING:\n\tParmeter {key} not found in file. Nothing changed, moving on.")

    # If general_metals dict not empty then change vals
    if general_metals_to_edit:
        keys_needed = ["Rdc", "Rrf", "Xdc", "Ls"]
        for metal_name, vals in general_metals_to_edit.items():
            # check that the correct keys are in the dict
            if not all(key in vals for key in keys_needed):
                print(
                    f'The "{metal_name}" general metal to edit does not have the correct keys. Keys needed are {keys_needed}. The keys present are {list(vals.keys())}.'
                )
                continue

            full_pattern = re.compile(r'MET "' + metal_name + r'" \d+ SUP( [+\-]?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+\-]?\d+)?){4}')
            match = full_pattern.search(contents)
            if match:
                metal_match_string = match.group()

                Rdc = vals["Rdc"]
                Rrf = vals["Rrf"]
                Xdc = vals["Xdc"]
                Ls = vals["Ls"]

                values_pattern = re.compile(r"( [+\-]?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+\-]?\d+)?){4}")
                values_to_put_in = f" {Rdc} {Rrf} {Xdc} {Ls}"

                new_metal_string = values_pattern.sub(values_to_put_in, metal_match_string)

                contents = full_pattern.sub(new_metal_string, contents)
            else:
                print(f"WARNING:\n\tGeneral metal {metal_name} not found in file. Nothing changed, moving on.")

    # If sweeps dict not empty then change vals
    if adaptive_sweeps_to_edit:
        keys_needed = ["sweep_min", "sweep_max", "target_freqs"]
        # check that the correct keys are in the dict and if so do replacement
        if all(key in adaptive_sweeps_to_edit for key in keys_needed):
            full_pattern = re.compile(r"FREQ \w+ AY ABS_ENTRY( (-)?[0-9]\d*(\.\d+)?){4}")
            match = full_pattern.search(contents)
            if match:
                abs_sweep_match_string = match.group()

                new_sweep_min = adaptive_sweeps_to_edit["sweep_min"]
                new_sweep_max = adaptive_sweeps_to_edit["sweep_max"]
                target_freqs = adaptive_sweeps_to_edit["target_freqs"]

                sweep_min_max_pattern = re.compile(r"ABS_ENTRY( (-)?[0-9]\d*(\.\d+)?){2}")
                sweep_min_max_replacement = f"ABS_ENTRY {new_sweep_min} {new_sweep_max}"
                new_abs_sweep_string = sweep_min_max_pattern.sub(sweep_min_max_replacement, abs_sweep_match_string)

                target_freqs_pattern = re.compile(r"(\d+)\D*$")
                target_freqs_replacement = f"{target_freqs}"
                new_abs_sweep_string = target_freqs_pattern.sub(target_freqs_replacement, new_abs_sweep_string)

                contents = full_pattern.sub(new_abs_sweep_string, contents)
            else:
                print("WARNING:\n\tCan't find an adaptive sweep to alter. Nothing changed, moving on.")

        else:
            print(
                f"The adaptive sweeps to edit does not have the correct keys. Keys needed are {keys_needed}. The keys present are {list(adaptive_sweeps_to_edit.keys())}."
            )

    # If sweeps dict not empty then change vals
    if linear_sweeps_to_edit:
        keys_needed = ["sweep_min", "sweep_max", "step_size"]
        # check that the correct keys are in the dict and if so do replacement
        if all(key in linear_sweeps_to_edit for key in keys_needed):
            full_pattern = re.compile(r"FREQ \w+ \w+ SWEEP( (-)?[0-9]\d*(\.\d+)?){3}")
            match = full_pattern.search(contents)
            if match:
                linear_sweep_match_string = match.group()

                new_sweep_min = linear_sweeps_to_edit["sweep_min"]
                new_sweep_max = linear_sweeps_to_edit["sweep_max"]
                step_size = linear_sweeps_to_edit["step_size"]

                values_pattern = re.compile(r"( (-)?[0-9]\d*(\.\d+)?){3}")
                values_to_put_in = f" {new_sweep_min} {new_sweep_max} {step_size}"
                new_linear_sweep_string = values_pattern.sub(values_to_put_in, linear_sweep_match_string)

                contents = full_pattern.sub(new_linear_sweep_string, contents)
            else:
                print("WARNING:\n\tCan't find an linear sweep to alter. Nothing changed, moving on.")

        else:
            print(
                f"The linear sweeps to edit does not have the correct keys. Keys needed are {keys_needed}. The keys present are {list(linear_sweeps_to_edit.keys())}."
            )

    with open(final_file, "w") as f:
        f.write(contents)

    print(f"Written file to drive:\n\tname: {output_filename}\n\tpath: {output_file_path}")
    return



import filecmp
import inspect
import os
import shutil

import pytest

from sonnetsuiteshelper import file_generation

"""
tests - file_generation.py.
    generate_file_like function testing.
        [X] - base file is not present
        [X] - output directory doesnt exist

        [X] - make file with no alterations
        [X] - no .son file extentions

        [X] - make file with 1 param changed
        [X] - make file with many params changed
        [X] - make file with params not in file changed

        [X] - make file with 1 general metal changed
        [X] - make file with many general metals changed
        [X] - make file with general metals not in file changed
        [X] - make file with many general metals with incorrect keys

        [X] - make file with many general metals and many params changed

        [X] - make file with adaptive sweep changed

        [X] - make file with linear sweep changed
"""

desired_output_files_path = r"tests/test_files/test_desired_output_files"
generated_output_file_path = r"tests/test_files/test_output_files"

base_filename_to_edit = r"general_base_file.son"
base_file_path_to_edit = r"tests/test_files/test_base_files"


def test_base_file_not_present() -> None:
    base_filename = "not_a_file.son"
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {}
    sweeps_to_edit = {}

    # check that is raises a FileNotFoundError
    with pytest.raises(FileNotFoundError):
        file_generation.generate_file_like(
            base_filename=base_filename,
            output_filename=output_filename,
            base_file_path=base_file_path,
            output_file_path=output_file_path,
            output_filename_prefix=output_filename_prefix,
            output_filename_suffix=output_filename_suffix,
            params_to_edit=params_to_edit,
            general_metals_to_edit=general_metals_to_edit,
            adaptive_sweeps_to_edit=sweeps_to_edit,
        )


def test_output_dir_doesnt_exist() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path + r"/new_folder_created"
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {}
    sweeps_to_edit = {}

    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    if os.path.isfile(generated_file):
        shutil.rmtree(output_file_path)

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__no_alterations() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {}
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__no_son_extentions() -> None:
    base_filename = base_filename_to_edit[:-4]
    output_filename = inspect.stack()[0][3] + ".son"  # current function name + ".son"
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {}
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename[:-4]}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__1_param_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {"COUPLER": 222}
    general_metals_to_edit = {}
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__many_params_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {
        "IDC_ARM_10": 1100,
        "IDC_ARM_11": 1100,
        "IDC_ARM_12": 1100,
        "IDC_ARM_13": 1100,
        "IDC_ARM_14": 1100,
        "IDC_ARM_15": 1100,
        "IDC_ARM_16": 1100,
        "IDC_ARM_17": 1100,
        "IDC_ARM_18": 1100,
        "IDC_ARM_19": 1100,
    }
    general_metals_to_edit = {}
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__params_not_in_file_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {
        "NOT_A_VARIABLE": 1500,
        "TESTING": 100,
        "idc_arm_10": 1234,
    }
    general_metals_to_edit = {}
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__1_general_metal_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {"Al": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8}}
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__many_general_metals_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {
        "Al": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        ".mets": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        "Nb": {"Rdc": 1e-08, "Rrf": 1e-18, "Xdc": 0, "Ls": 0.003},
        "AL_alteredLK": {"Rdc": 0, "Rrf": 0, "Xdc": 0, "Ls": 10.3},
    }
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__many_general_metals_not_in_file() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {
        "Not a metal": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        "still_not_a_metal": {"Rdc": 0, "Rrf": 0, "Xdc": 0, "Ls": 10.3},
    }
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__many_general_metals_with_incorrect_keys() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {
        "Al": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        ".mets": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        "Nb": {"RdWc": 1e-08, "Rrf": 1e-18, "Xdc": 0, "Ls": 0.003},
        "AL_alteredLK": {
            "Rdc": 0,
            "Rrf": 0,
            "Xdc": 0,
            "Ls": 10.3,
            "Another_irelevant_key": 1,
        },
    }
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__many_general_metals_and_many_params_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {"COUPLER": 100, "IDC_ARM_20": 400}
    general_metals_to_edit = {
        "Al": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        ".mets": {"Rdc": 0.5, "Rrf": 0.6, "Xdc": 0.7, "Ls": 0.8},
        "Nb": {"Rdc": 1e-08, "Rrf": 1e-18, "Xdc": 0, "Ls": 0.003},
        "AL_alteredLK": {"Rdc": 0, "Rrf": 0, "Xdc": 0, "Ls": 10.3},
    }
    sweeps_to_edit = {}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__adaptive_sweep_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {}
    sweeps_to_edit = {"sweep_min": 8.5, "sweep_max": 10.5, "target_freqs": 500}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        adaptive_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output


def test_generate_file_like__linear_sweep_changed() -> None:
    base_filename = base_filename_to_edit
    output_filename = inspect.stack()[0][3]  # current function name
    base_file_path = base_file_path_to_edit
    output_file_path = generated_output_file_path
    output_filename_prefix = ""
    output_filename_suffix = ""
    params_to_edit = {}
    general_metals_to_edit = {}
    sweeps_to_edit = {"sweep_min": 0.2, "sweep_max": 15.3, "step_size": 0.01}

    file_generation.generate_file_like(
        base_filename=base_filename,
        output_filename=output_filename,
        base_file_path=base_file_path,
        output_file_path=output_file_path,
        output_filename_prefix=output_filename_prefix,
        output_filename_suffix=output_filename_suffix,
        params_to_edit=params_to_edit,
        general_metals_to_edit=general_metals_to_edit,
        linear_sweeps_to_edit=sweeps_to_edit,
    )

    # compare the output to the desired output.
    generated_filename = f"{output_filename_prefix}{output_filename}{output_filename_suffix}.son"
    generated_file = os.path.join(output_file_path, generated_filename)

    desired_output_file = os.path.join(desired_output_files_path, f"{output_filename}.son")

    generated_file_matches_desired_output = filecmp.cmp(generated_file, desired_output_file)

    assert generated_file_matches_desired_output



#!/usr/bin/env python
import re
from pathlib import Path

from setuptools import find_packages, setup


def read(*names, **kwargs):
    with Path(__file__).parent.joinpath(*names).open(encoding=kwargs.get("encoding", "utf8")) as fh:
        return fh.read()


setup(
    name="sonnetsuiteshelper",
    version="0.4.6",
    license="LGPL-3.0-or-later",
    description="A package to help with Sonnet Suites.",
    long_description="{}\n{}".format(
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub("", read("README.rst")),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    author="Alan Manning",
    author_email="Alan_Manning@Live.co.uk",
    url="https://github.com/Alan-Manning/python-sonnetsuiteshelper",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[path.stem for path in Path("src").glob("*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Utilities",
    ],
    project_urls={
        "Documentation": "https://python-sonnetsuiteshelper.readthedocs.io/",
        "Changelog": "https://python-sonnetsuiteshelper.readthedocs.io/en/latest/changelog.html",
        "Issue Tracker": "https://github.com/Alan-Manning/python-sonnetsuiteshelper/issues",
    },
    keywords=[
        "sonnet",
        "sonnetsuites",
        ".son",
    ],
    python_requires=">=3.10",
    install_requires=[
        "numpy",
        "pandas",
        "scipy",
        "matplotlib",
        "pyyaml",
    ],
    extras_require={},
    entry_points={
        "console_scripts": [
            "sonnetsuiteshelper = sonnetsuiteshelper.cli:main",
        ]
    },
)
